/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
  tranquille yéyé et les autres avant aussi
	utiliser avec ESP01S
  balise en mode station, sleep mode entre les trames.
 */

//commenter la ligne suivante si on veut utiliser hdop plutôt que pdop dans la précision gps:
#define PDOP

#include <ESP8266WiFi.h>
#include <SoftwareSerial.h>  

extern "C" 
{
	#include "user_interface.h"
	int wifi_send_pkt_freedom(uint8 *buf, int len, bool sys_seq);
	int  wifi_set_user_fixed_rate(uint8  enable_mask,  uint8 rate);
	bool  wifi_set_user_rate_limit(uint8  mode, uint8 ifidx,  uint8 max, uint8 min);
	bool  wifi_set_user_limit_rate_mask(uint8 enable_mask);
	int  wifi_set_user_sup_rate(uint8  min,  uint8 max);
}

uint32_t sleep_time_in_ms = 10000;
void  Reveiller_Balise(void)
{
    wifi_fpm_close();            // disable force sleep function
    wifi_set_opmode(STATION_MODE);              //  set station mode
}

void  Endormir_Balise()
{   
    wifi_set_opmode(NULL_MODE);                  // set WiFi  mode  to  null  mode.
    wifi_fpm_set_sleep_type(LIGHT_SLEEP_T);          // light sleep
    wifi_fpm_open();                             // enable  force sleep
    wifi_fpm_set_wakeup_cb(Reveiller_Balise); //  Set wakeup  callback  
    wifi_fpm_do_sleep(sleep_time_in_ms*1000);     
    delay(50);      //le délai permet d'entrer dans le mode sleep, car il s'active sur une inactivité uniquement
}

#define LED 2 // led bleue (pour le esp-01s c'est la pin 2 (si esp-01 c'est pin 1, conflit avec le serial, donc mettre ici genre 4 pour désactiver cette fonction)
#define ON LOW
#define OFF HIGH
#define GPS_RX_PIN 0       // Brancher le fil Tx du GPS
#define GPS_TX_PIN 2       // Rx gps, pas utilisé
SoftwareSerial Serial2(GPS_RX_PIN,GPS_TX_PIN);

#define GPS_BAUD_RATE 9600

#include <TinyGPS++.h>
#include "droneID_FR.h"

TinyGPSPlus gps;
TinyGPSCustom pdop(gps, "GPGSA", 15); // $GPGSA sentence, 15th element
droneIDFR drone_idfr;


/********************************************************************************************************************
 * MODIFIEZ LES VALEURS ICI
 ********************************************************************************************************************
  * CHANGEZ l'ID du drone par celui que Alphatango vous a fourni (Trigramme + Modèle + numéro série) !
  ILLEGAL_DRONE_APPELEZ_POLICE17
  */
const char ssid[] = "BALISE_DGAC-XAV"; //il en faut un pour mettre dans la trame beacon, on met ce qu'on veut c'est pas visible ailleurs...
//ILLEGAL_DRONE_APPELEZ_POLICE17

const char drone_id[] = "ILLEGAL_DRONE_APPELEZ_POLICE17";

/********************************************************************************************************************
 * XAV
 ********************************************************************************************************************/
 // beacon frame definition
static constexpr uint16_t MAX_BEACON_SIZE = 40 + 32 + droneIDFR::FRAME_PAYLOAD_LEN_MAX;  // default beaconPacket size + max ssid size + max drone id frame size
uint8_t beaconPacket[MAX_BEACON_SIZE] = {
        0x80, 0x00,                         // 0-1: Frame Control
        0x00, 0x00,                         // 2-3: Duration
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff,       // 4-9: Destination address (broadcast)
        0x24, 0x62, 0xab, 0xdd, 0xb0, 0xbd,       // 10-15: Source address FAKE  // TODO should bet set manually
        0x24, 0x62, 0xab, 0xdd, 0xb0, 0xbd,       // 16-21: Source address FAKE
        0x00, 0x00,                         // 22-23: Sequence / fragment number (done by the SDK)
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, // 24-31: Timestamp (GETS OVERWRITTEN TO 0 BY HARDWARE)
        0xB8, 0x0B,                         // 32-33: Beacon interval: set to 3s == 3000TU== BB8, bytes in reverse order  // TODO: manually set it
        0x21, 0x04,                         // 34-35: Capability info
        0x03, 0x01, 0x06,                   // 36-38: DS Parameter set, current channel 6 (= 0x06), // TODO: manually set it
        0x00, 0x20,                             // 39-40: SSID parameter set, 0x20:maxlength:content
                                                        // 41-XX: SSID (max 32)
};

// activer ou non le DEBUG
boolean DEBUG = false;       

const int ClignoteErreur = 25;
boolean IsErreur = false;
// si erreur sur gps, on pause un certain temps et on retente, pas la peine d'interroger toutes les millisecondes...
// mis à 0 car sinon problème de lecture du gps avec les délais...
const int delayErreur = 0;
double codeinfo=0;

// dès que la position home a été mise, on passe à true et on y reste
bool has_set_home = false;
double home_alt = 0;
//les positions précédentes:
double lat_prev = 0;
double lon_prev = 0;
int16_t alt_prev = 0;
int16_t dir_prev = 0;
double vit_prev = 0;

uint64_t beaconSec = 0;

// lecture des données du GPS, la led s'allume puis s'éteint brièvement quand on est dans cette fonction 
//du coup elle reste bien allumée car on y passe souvent...
static void RecupDonneesGPS()
{
  // si on est en erreur, on laisse la led éteinte
  // sinon on l'allume
	if (!IsErreur) digitalWrite(LED, ON);
  //récup des données gps
	while (Serial2.available()) { gps.encode(Serial2.read()); } 
  // on n'écrit rien ça sert à rien, même en débug
	//if (DEBUG) Serial.println("... Récupération données du GPS ... ");
}

static void EnvoiTrame(double lat_actu,double lon_actu, int16_t alt_actu, int16_t dir_actu, double vit_actu)
{
    //on définit le code info dans la trame:
    drone_idfr.set_codeinfo(codeinfo);
    //on envoie les données à la biblio pour fomattage:
    drone_idfr.set_current_position(lat_actu, lon_actu, alt_actu);
    drone_idfr.set_heading(dir_actu);
    drone_idfr.set_ground_speed(vit_actu);
    drone_idfr.set_heigth(alt_actu - home_alt);
    
     /**
     * On regarde s'il est temps d'envoyer la trame d'identification drone: soit toutes les 3s soit si le drones s'est déplacé de 30m en moins de 3s.
     *  - soit toutes les 3s,
     *  - soit si le drone s'est déplacé de 30m en moins de 30s soit 10m/s ou 36km/h,
     *  - uniquement si la position Home est déjà définie,
     *  - et dans le cas où les données GPS sont nouvelles.
     */
     
    // on envoie une trame dès qu'il est temps d'en envoyer une
    if (drone_idfr.time_to_send())
    {
      // si aucune erreur et que la trame est la bonne:
      if (codeinfo==9)
      {
        // LED allumée en fonctionnement normal,
        // deux brefs clignotements pendant l'envoi de la trame
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
      }
        //On commence par renseigner le ssid du wifi dans la trame
        // write new SSID into beacon frame
        const size_t ssid_size = (sizeof(ssid)/sizeof(*ssid)) - 1; // remove trailling null termination
        beaconPacket[40] = ssid_size;  // set size
        memcpy(&beaconPacket[41], ssid, ssid_size); // set ssid
        const uint8_t header_size = 41 + ssid_size;  //TODO: remove 41 for a marker
        
         //On génère la trame wifi avec l'identification
        const uint8_t to_send = drone_idfr.generate_beacon_frame(beaconPacket, header_size);  // override the null termination

        // Ici on ecrit sur le port USB les données GPS pour visualisation seulement.
          Serial.println("Temps UTC: " + String(gps.date.day()) + "/" + String(gps.date.month()) + "/"  + String(gps.date.year()) + " " + String(gps.time.hour()) + ":" + String(gps.time.minute()) + ":"  + String(gps.time.second()));
          Serial.println("Position de départ: " + drone_idfr.get_home_position());
          Serial.println("Position actuelle : " + drone_idfr.get_position_actuelle());
          Serial.println("Nb de satellites: " + String(gps.satellites.value()));
          #ifdef PDOP 
			Serial.println("pdop: " + String(pdop.value()));
		  #else
			Serial.println("hdop: " + String(gps.hdop.value()));
		  #endif
          Serial.println("Code info: " + String(codeinfo));
          Serial.println("--------------------------");

          //allume wifi
          Reveiller_Balise();
          //envoi trame
          wifi_send_pkt_freedom(beaconPacket, to_send, 0);
          //éteindre wifi
          Endormir_Balise();
        
        //On reset la condition d'envoi
        drone_idfr.set_last_send();
        
      // si aucune erreur et que la trame est la bonne:
      if (codeinfo==9)
      {
        // LED allumée en fonctionnement normal,
        // deux brefs clignotements pendant l'envoi de la trame
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
      }
    }
}

// Ensure the drone_id is max 30 letters
static_assert((sizeof(drone_id)/sizeof(*drone_id))<=31, "Drone ID should be less that 30 letters !");  // 30 lettres + null termination

/**
 * Phase de configuration.
 */
void setup()
{
    Serial.begin(115200);
    delay(5000);
    
    //initialisation du GPS. Bien penser à croiser Rx et Tx entre ESPxx et le GPS!
      Serial2.begin(GPS_BAUD_RATE);
    
    //définition de la led bleue
    pinMode(LED, OUTPUT);

  //init wifi
    Serial.println("--------------------------");
    Serial.println("Mise en place du wifi:");

      system_phy_set_powerup_option(3);  // Do full RF calibration on power-up
      system_phy_set_rfoption(2); // No RF calibration after Deep-sleep wakeup
      //power tx:
      system_phy_set_max_tpw(82);
      //wifi g
      wifi_set_phy_mode(PHY_MODE_11G);
      //canal
      wifi_set_channel(6);
      // mode sta sinon avec 0 aucune trame émise (c'est plus pour mode sniffeur)
      wifi_set_opmode(STATION_MODE); //STATION_MODE SOFTAP_MODE

      //définition de la vitesse d'envoi (rate)
      wifi_set_user_fixed_rate(FIXED_RATE_MASK_ALL,  PHY_RATE_6);
      wifi_set_user_rate_limit(RC_LIMIT_11G,  1,  RATE_11G_G6M, RATE_11G_G6M);
      wifi_set_user_limit_rate_mask(LIMIT_RATE_MASK_ALL);
      wifi_set_user_sup_rate(RATE_11G6M,  RATE_11G6M);
      
      //largeur de bande
      //par défaut 20 si wifi G?
   
    //affichage des infos
    Serial.print("Nom de la balise: ");Serial.println(drone_id);

  // définition de l'id de la balise:
    drone_idfr.set_drone_id(drone_id);
    delay(2000);
    //Serial.print("phy: ");Serial.println(String(wifi_get_phy_mode()));
    Serial.println("--------------------------");

    //éteindre balise:
    Endormir_Balise();
} // fin setup

/**
 * Début du code principal. C'est une boucle infinie.
 */
void loop()
{
// on attend un peu que tout s'init bien (5s à partir du démarrage de la balise, en général quand on est ici c'est déjà passé les 10s)
// ne s'exec qu'une seule fois au démarrage de la balise ce délai:
	while (millis() < 10000)
	{
		digitalWrite(LED, ON);
		Serial.print(".");
		delay(100);
		digitalWrite(LED, OFF);
		delay(100);
	}

// on récupère les données du GPS
	RecupDonneesGPS();

	if (gps.charsProcessed() < 10)
	{
		if (DEBUG) Serial.println("GPS non détecté, aucune donnée GPS reçue");
    digitalWrite(LED, OFF);
    IsErreur = true;
		if (DEBUG) Serial.println("--------------------------");
    delay(delayErreur);
    //envoyer la trame
    codeinfo= 1;
    EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
		return;
	}

	if (!gps.location.isValid())
	{
		if (DEBUG) Serial.println("Positionnement du GPS non valide (fix en cours?)");
    digitalWrite(LED, OFF);
    IsErreur = true;
		if (DEBUG) Serial.println("--------------------------");
    delay(delayErreur);
    //envoyer la trame
    codeinfo= 2;
    EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
    return;
	}
	if (gps.location.age()> 3000)
	{
		if (DEBUG) Serial.println("Fix du GPS perdu...");
    digitalWrite(LED, OFF);
    IsErreur = true;
		if (DEBUG) Serial.println("--------------------------");
    delay(delayErreur);
    //envoyer la trame
    codeinfo= 3;
    EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
    return;
	}

// test si précision ok, sinon on remonte au début afin de ne pas encore fixer la home position
    if (gps.satellites.value() < 4)
    {
       digitalWrite(LED, OFF);
       IsErreur = true;
       if (DEBUG) Serial.println("La précision du GPS n'est pas bonne");
       if (DEBUG) Serial.println("Nb de satellites (doit être supérieur à 3): " + String(gps.satellites.value()));
       if (DEBUG) Serial.println("--------------------------");
       delay(delayErreur);
       //envoyer la trame
       codeinfo= 4;
       EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
       return;
    }
	#ifdef PDOP 
		if ((String(pdop.value()).toDouble() == 0.0) || (String(pdop.value())=="") || (String(pdop.value()).toDouble() > 5.0))
	#else
		if (gps.hdop.hdop() > 5.0)
	#endif
    {
		digitalWrite(LED, OFF);
		IsErreur = true;
		if (DEBUG) Serial.println("La précision du GPS n'est pas bonne");
		#ifdef PDOP 
				if (DEBUG) Serial.println("Précision 3D (doit être inférieure à 5.0): " + String(pdop.value()));
		#else
				if (DEBUG) Serial.println("Précision 3D (doit être inférieure à 5.0): " + String(gps.hdop.hdop()));
		#endif
       if (DEBUG) Serial.println("--------------------------");
       delay(delayErreur);
       //envoyer la trame
       codeinfo= 5;
       EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
       return;
    }
    if (gps.altitude.meters() == 0.0)
    {
        digitalWrite(LED, OFF);
        IsErreur = true;
        if (DEBUG) Serial.println("La précision du GPS n'est pas bonne");
        if (DEBUG) Serial.println("Altitude non récupérée: " + String(gps.altitude.meters()));
        if (DEBUG) Serial.println("--------------------------");
        delay(delayErreur);
        //envoyer la trame
        codeinfo= 6;
       EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
        return;
    }

     // ici on n'est plus en erreur
    IsErreur = false;
    
    // On renseigne une première et unique fois le point de démarrage dès que la précision est satisfaisante
    //
    if (!has_set_home)
    {
        Serial.println("Stockage de la position de départ");
        Serial.println("Nb de satellites (doit être supérieur à 3): " + String(gps.satellites.value()));
		#ifdef PDOP
			Serial.println("Précision 3D (doit être inférieure à 5.0): " + String(pdop.value()));
		#else
			Serial.println("Précision 3D (doit être inférieure à 5.0): " + String(gps.hdop.hdop()));
		#endif
        home_alt = gps.altitude.meters();
        //on fixe la position home:
        drone_idfr.set_home_position(gps.location.lat(), gps.location.lng(), home_alt);
  		  has_set_home = true;
        
        //led bleue rythme rock'n'roll, tout est ok!
        int compteur = 20;
        while(compteur > 0)
        {
          digitalWrite(LED, ON);
          delay(50);
          digitalWrite(LED, OFF);
          delay(50);
          compteur--;
        }
        // on lit la position home pour controle
        Serial.println("Position de départ: " + drone_idfr.get_home_position());
        Serial.println("Stockage de la position de départ OK");
        Serial.println("--------------------------");
        //on envoie une trame
        codeinfo= 7;
        EnvoiTrame(gps.location.lat(),gps.location.lng(), home_alt, gps.course.deg(), gps.speed.mps());
        return;
    }

    // si ici, pas d'erreur donc:
    codeinfo = 9;
    //envoi de la trame ici
    EnvoiTrame(gps.location.lat(),gps.location.lng(), gps.altitude.meters(), gps.course.deg(), gps.speed.mps());
    //on stocke les dernières positions
    lat_prev = gps.location.lat();
    lon_prev = gps.location.lng();
    alt_prev = gps.altitude.meters();
    dir_prev = gps.course.deg();
    vit_prev = gps.speed.mps();
}
